from nameparser import HumanName

from whoswho.config import UNIQUE_SUFFIXES, MALE_TITLES, FEMALE_TITLES
from whoswho.utils import make_ascii, strip_punctuation, compare_name_component


class Name(object):

    def __init__(self, fullname):
        ascii_name = make_ascii(unicode(fullname))
        self.name = HumanName(ascii_name)

        # Format after parsing to preserve parsing logic
        self.name.title = strip_punctuation(self.name.title)
        self.name.first = strip_punctuation(self.name.first)
        self.name.middle = strip_punctuation(self.name.middle)
        self.name.last = strip_punctuation(self.name.last)
        self.name.suffix = strip_punctuation(self.name.suffix)

    def deep_compare(self, other, settings):
        """
        Compares each field of the name one at a time to see if they match.
        Each name field has context-specific comparison logic.

        :param Name other: other Name for comparison
        :return bool: whether the two names are compatible
        """

        if not self._is_compatible_with(other):
            return False

        first, middle, last = self._compare_components(other, settings)

        return first and middle and last

    def ratio_deep_compare(self, other, settings):
        """
        Compares each field of the name one at a time to see if they match.
        Each name field has context-specific comparison logic.

        :param Name other: other Name for comparison
        :return int: ratio fuzzy match (out of 100)
        """

        if not self._is_compatible_with(other):
            return 0

        first, middle, last = self._compare_components(other, settings, True)
        f_weight, m_weight, l_weight = self._determine_weights(other, settings)
        total_weight = f_weight + m_weight + l_weight

        result = (
            first * f_weight +
            middle * m_weight +
            last * l_weight
        ) / total_weight

        return result

    def _is_compatible_with(self, other):
        """
        Return True if names are not incompatible.

        This checks that the gender of titles and compatibility of suffixes
        """

        title = self._compare_title(other)
        suffix = self._compare_suffix(other)

        return title and suffix

    def _compare_title(self, other):
        """Return False if titles have different gender associations"""

        # If title is omitted, assume a match
        if not self.name.title or not other.name.title:
            return True

        titles = set(self.name.title_list + other.name.title_list)

        return not (titles & MALE_TITLES and titles & FEMALE_TITLES)

    def _compare_suffix(self, other):
        """Return false if suffixes are mutually exclusive"""

        # If suffix is omitted, assume a match
        if not self.name.suffix or not other.name.suffix:
            return True

        # Check if more than one unique suffix
        suffix_list = self.name.suffix_list + other.name.suffix_list
        unique_suffixes = {s for s in suffix_list if s in UNIQUE_SUFFIXES}

        return len(unique_suffixes) < 2

    def _compare_components(self, other, settings, ratio=False):
        """Return comparison of first, middle, and last components"""

        first = compare_name_component(
            self.name.first_list,
            other.name.first_list,
            settings['first'],
            ratio,
        )
        middle = compare_name_component(
            self.name.middle_list,
            other.name.middle_list,
            settings['middle'],
            ratio,
        )
        last = compare_name_component(
            self.name.last_list,
            other.name.last_list,
            settings['last'],
            ratio,
        )

        return first, middle, last

    def _determine_weights(self, other, settings):
        """
        Return weights of name components based on whether or not they were
        omitted
        """

        # TODO: Reduce weight for matches by prefix or initials

        first_is_used = settings['first']['required'] or \
            self.name.first and other.name.first
        first_weight = settings['first']['weight'] if first_is_used else 0

        middle_is_used = settings['middle']['required'] or \
            self.name.middle and other.name.middle
        middle_weight = settings['middle']['weight'] if middle_is_used else 0

        last_is_used = settings['last']['required'] or \
            self.name.last and other.name.last
        last_weight = settings['last']['weight'] if last_is_used else 0

        return first_weight, middle_weight, last_weight



from .settings import *
from .suffixes import *
from .titles import *

# Punctuation that may or may not appear in equivalent names
STRIPPED_CHARACTERS = dict.fromkeys(map(ord, ".'"), None)

SETTINGS = {
    'default': DEFAULT_SETTINGS,
    'strict': STRICT_SETTINGS,
    'lenient': LENIENT_SETTINGS
}


# -*- coding: utf-8 -*-
from __future__ import unicode_literals


UNIQUE_SUFFIXES = {
    'jr',
    'jnr',  # TODO: compatible with jr
    'sr',
    'snr',  # TODO: compatible with sr
    '2',
    'i',
    'ii',
    'iii',
    'iv',
    'v',
}
"""
A set of suffixes that should be mutually exclusive, which is a
subset of nameparser's set of suffixes.

Suffixes that are actually titles are included as suffixes by nameparser,
but differing titles do not necessarily indicate different people.
"""


MISC_SUFFIXES = {
    'esq',
    'esquire',
    'clu',
    'chfc',
    'cfp',
    'md',
    'phd',
    'mp',
    'qc',
    'dmd',
    'do',
    'dds',
    'dpm',
}
"""
Non-unique suffixes kept to ensure the set is up to date with nameparser
"""



from setuptools import setup, find_packages

REQUIREMENTS = ['nameparser', 'fuzzywuzzy']
TEST_REQUIREMENTS = ['nose']

setup(
    name='whoswho',
    packages=find_packages(),
    version='0.1.1',
    description='A simple python library for determining whether '
                'two names describe the same person.',
    author='Robert Liebowitz',
    author_email='rliebz@gmail.com',
    url='https://www.github.com/rliebz/whoswho',
    install_requires=REQUIREMENTS,
    tests_require=TEST_REQUIREMENTS,
    classifiers=(
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
    )
)


DEFAULT_SETTINGS = {
    'first': {
        'required': True,
        'allow_initials': True,
        'allow_prefix': True,
        'weight': 3,
    },
    'middle': {
        'required': False,
        'allow_initials': True,
        'allow_prefix': False,
        'weight': 2,
    },
    'last': {
        'required': True,
        'allow_initials': False,
        'allow_prefix': False,
        'weight': 5,
    },
}

STRICT_SETTINGS = {
    'first': {
        'required': True,
        'allow_initials': False,
        'allow_prefix': False,
        'weight': 3,
    },
    'middle': {
        'required': False,
        'allow_initials': False,
        'allow_prefix': False,
        'weight': 2,
    },
    'last': {
        'required': True,
        'allow_initials': False,
        'allow_prefix': False,
        'weight': 5,
    },
}

LENIENT_SETTINGS = {
    'first': {
        'required': False,
        'allow_initials': True,
        'allow_prefix': True,
        'weight': 3,
    },
    'middle': {
        'required': False,
        'allow_initials': True,
        'allow_prefix': True,
        'weight': 2,
    },
    'last': {
        'required': True,
        'allow_initials': True,
        'allow_prefix': False,
        'weight': 5,
    },
}



# -*- coding: utf-8 -*-
import nose
from nose.tools import *
import unittest

from whoswho import who, config, utils

from nameparser.config.titles import TITLES as NAMEPARSER_TITLES
from nameparser.config.suffixes import SUFFIXES as NAMEPARSER_SUFFIXES


class TestFullNames(unittest.TestCase):

    def setUp(self):
        self.name = 'Robert Evan Liebowitz'

    def test_unicode(self):
        assert_true(who.match(self.name, u'attaché Robert Evan Liebowitz'))
        assert_true(who.match(self.name, u'Rōbért Èvān Lîęböwitz'))
        assert_false(who.match(self.name, u'Rōbért Èvān Lęîböwitz'))

    def test_name_and_initials(self):
        assert_true(who.match(self.name, 'R. Evan Liebowitz'))
        assert_true(who.match(self.name, 'Robert E. Liebowitz'))
        assert_true(who.match(self.name, 'R. E. Liebowitz'))

    def test_different_number_initials(self):
        assert_true(who.match(self.name, 'Robert Liebowitz'))
        assert_true(who.match(self.name, 'R. Liebowitz'))
        assert_false(who.match(self.name, 'Robert E. E. Liebowitz'))
        assert_false(who.match(self.name, 'R. E. E. Liebowitz'))
        assert_true(who.match('R.E.E. Liebowitz', 'R. E. E. Liebowitz'))

    def test_different_initials(self):
        assert_false(who.match(self.name, 'E. R. Liebowitz'))
        assert_false(who.match(self.name, 'E. Liebowitz'))
        assert_false(who.match(self.name, 'R. V. Liebowitz'))
        assert_false(who.match(self.name, 'O. E. Liebowitz'))

    def test_short_names(self):
        assert_true(who.match(self.name, 'Rob Liebowitz'))
        # TODO: Should these be true?
        assert_false(who.match(self.name, 'Bert Liebowitz'))
        assert_false(who.match(self.name, 'Robbie Liebowitz'))

    def test_suffixes(self):
        name = 'Robert Liebowitz Jr'
        assert_true(who.match(name, 'Robert Liebowitz'))
        assert_true(who.match(name, 'Robert Liebowitz Jr'))
        assert_true(who.match(name, 'Robert Liebowitz, PhD'))
        assert_false(who.match(name, 'Robert Liebowitz, Sr'))
        assert_false(who.match(name, 'Robert Liebowitz, Sr, PhD'))
        assert_true(who.match(name, 'Robert Liebowitz, Jr, PhD'))

    def test_titles(self):
        name = 'Mr. Robert Liebowitz'
        assert_true(who.match(name, 'Robert Liebowitz'))
        assert_true(who.match(name, 'Sir Robert Liebowitz'))
        assert_true(who.match(name, 'Dr. Robert Liebowitz'))
        assert_false(who.match(name, 'Mrs. Robert Liebowitz'))


class TestConfig(unittest.TestCase):

    def test_titles_all_defined(self):
        """
        Check if list of titles is up to date with nameparser
        """
        all_titles = (
            config.MALE_TITLES |
            config.FEMALE_TITLES |
            config.GENDERLESS_TITLES
        )
        assert_equal(all_titles, NAMEPARSER_TITLES)

    def test_suffixes_all_defined(self):
        """
        Check if list of suffixes is up to date with nameparser
        """
        all_suffixes = (
            config.UNIQUE_SUFFIXES |
            config.MISC_SUFFIXES
        )
        assert_equal(all_suffixes, NAMEPARSER_SUFFIXES)


class TestUtils(unittest.TestCase):

    def test_equate_prefix_to_name(self):
        assert_true(utils.equate_prefix('r', 'robert'))
        assert_true(utils.equate_prefix('rob', 'robert'))
        assert_false(utils.equate_prefix('robbie', 'robert'))
        assert_false(utils.equate_prefix('bert', 'robert'))

    def test_make_ascii(self):
        assert_equal(
            utils.make_ascii(u"foo bar .,?;'!@#$%^&*()"),
            "foo bar .,?;'!@#$%^&*()"
        )
        assert_equal(
            utils.make_ascii(u'äèîõù'),
            'aeiou'
        )

    def test_strip_punctuation(self):
        assert_equal(
            utils.strip_punctuation(u'abcde aeiou'),
            u'abcde aeiou'
        )
        assert_equal(
            utils.strip_punctuation(u"abcde.' aeiou"),
            u'abcde aeiou'
        )

    def test_equate_nickname(self):
        assert_true(utils.equate_nickname('robert', 'rob'))
        assert_true(utils.equate_nickname('robert', 'robby'))
        assert_true(utils.equate_nickname('robert', 'robbie'))
        assert_true(utils.equate_nickname('robbie', 'robby'))
        assert_false(utils.equate_nickname('robert', 'robin'))
        assert_false(utils.equate_nickname('harold', 'harriet'))


if __name__ == '__main__':
    nose.main()


import re
import unicodedata

from fuzzywuzzy import fuzz

from whoswho.config import STRIPPED_CHARACTERS


def compare_name_component(list1, list2, settings, ratio=False):
    """
    Compare a list of names from a name component based on settings
    """
    if not list1[0] or not list2[0]:
        result = not settings['required']
        return result * 100 if ratio else result

    if len(list1) != len(list2):
        return False

    if ratio:
        result = 0
        for i, n1 in enumerate(list1):
            n2 = list2[i]

            # If initials don't match, result for this item is 0
            if (len(n1) == 1 or len(n2) == 1) and not equate_initial(n1, n2):
                continue

            if settings['allow_prefix']:
                result += 100 if equate_prefix(n1, n2) else fuzz.ratio(n1, n2)
            elif settings['allow_initials']:
                result += 100 if equate_initial(n1, n2) else fuzz.ratio(n1, n2)
            else:
                result += fuzz.ratio(n1, n2)

        result /= len(list1)

    else:
        result = True
        for i, n1 in enumerate(list1):
            n2 = list2[i]

            if settings['allow_prefix']:
                result &= equate_prefix(n1, n2)
            elif settings['allow_initials']:
                result &= equate_initial(n1, n2)
            else:
                result &= n1 == n2

    return result


def equate_initial(name1, name2):
    """
    Evaluates whether names match, or one name is the initial of the other
    """
    if len(name1) == 0 or len(name2) == 0:
        return False

    if len(name1) == 1 or len(name2) == 1:
        return name1[0] == name2[0]

    return name1 == name2


def equate_prefix(name1, name2):
    """
    Evaluates whether names match, or one name prefixes another
    """

    if len(name1) == 0 or len(name2) == 0:
        return False

    return name1.startswith(name2) or name2.startswith(name1)


def equate_nickname(name1, name2):
    """
    Evaluates whether names match based on common nickname patterns

    This is not currently used in any name comparison.
    """
    # Convert '-ie' and '-y' to the root name
    nickname_regex = r'(.)\1(y|ie)$'
    root_regex = r'\1'

    name1 = re.sub(nickname_regex, root_regex, name1)
    name2 = re.sub(nickname_regex, root_regex, name2)

    if equate_prefix(name1, name2):
        return True

    return False


def make_ascii(word):
    """
    Converts unicode-specific characters to their equivalent ascii
    """
    ascii_word = unicodedata.normalize('NFKD', word).encode('ascii', 'ignore')
    return unicode(ascii_word)


def strip_punctuation(word):
    """
    Strips punctuation from name and lower cases it
    """

    return word.translate(STRIPPED_CHARACTERS).lower()



from whoswho.model import Name
from whoswho.config import SETTINGS


def match(fullname1, fullname2, strictness='default'):
    """
    Takes two names and returns true if they describe the same person.

    :param string fullname1: first human name
    :param string fullname2: second human name
    :return bool: the names match
    """

    settings = SETTINGS[strictness]

    n1 = Name(fullname1)
    n2 = Name(fullname2)

    return n1.deep_compare(n2, settings)


def ratio(fullname1, fullname2, strictness='default'):
    """
    Takes two names and returns true if they describe the same person.
    Uses fuzzywuzzy's fuzzy matching on a per-field basis for names

    :param string fullname1: first human name
    :param string fullname2: second human name
    :return int: ratio fuzzy match (out of 100)
    """

    settings = SETTINGS[strictness]

    n1 = Name(fullname1)
    n2 = Name(fullname2)

    return n1.ratio_deep_compare(n2, settings)